"""Type flo.params.heatpumpwithbooststore, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

import gridworks.property_format as property_format
from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from gridworks.property_format import predicate_validator
from pydantic import BaseModel
from pydantic import validator

from gwatn.enums import DistributionTariff
from gwatn.enums import EmitterPumpFeedbackModel as PumpModel
from gwatn.enums import EnergySupplyType
from gwatn.enums import MixingValveFeedbackModel as MixingValveModel
from gwatn.enums import RecognizedCurrencyUnit
from gwatn.enums import RecognizedTemperatureUnit


class DistributionTariff000SchemaEnum:
    enum_name: str = "distribution.tariff.000"
    symbols: List[str] = [
        "00000000",
        "2127aba6",
        "ea5c675a",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class DistributionTariff000(StrEnum):
    Unknown = auto()
    VersantStorageHeatTariff = auto()
    VersantATariff = auto()

    @classmethod
    def default(cls) -> "DistributionTariff000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class DistributionTariffMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> DistributionTariff:
        if not DistributionTariff000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to DistributionTariff000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, DistributionTariff, DistributionTariff.default())

    @classmethod
    def local_to_type(cls, distribution_tariff: DistributionTariff) -> str:
        if not isinstance(distribution_tariff, DistributionTariff):
            raise SchemaError(
                f"{distribution_tariff} must be of type {DistributionTariff}"
            )
        versioned_enum = as_enum(
            distribution_tariff, DistributionTariff000, DistributionTariff000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, DistributionTariff000] = {
        "00000000": DistributionTariff000.Unknown,
        "2127aba6": DistributionTariff000.VersantStorageHeatTariff,
        "ea5c675a": DistributionTariff000.VersantATariff,
    }

    versioned_enum_to_type_dict: Dict[DistributionTariff000, str] = {
        DistributionTariff000.Unknown: "00000000",
        DistributionTariff000.VersantStorageHeatTariff: "2127aba6",
        DistributionTariff000.VersantATariff: "ea5c675a",
    }


class EmitterPumpFeedbackModel000SchemaEnum:
    enum_name: str = "emitter.pump.feedback.model.000"
    symbols: List[str] = [
        "00000000",
        "f6bde4fa",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class EmitterPumpFeedbackModel000(StrEnum):
    ConstantDeltaT = auto()
    ConstantGpm = auto()

    @classmethod
    def default(cls) -> "EmitterPumpFeedbackModel000":
        return cls.ConstantDeltaT

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class EmitterPumpFeedbackModelMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> PumpModel:
        if not EmitterPumpFeedbackModel000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to EmitterPumpFeedbackModel000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, PumpModel, PumpModel.default())

    @classmethod
    def local_to_type(cls, emitter_pump_feedback_model: PumpModel) -> str:
        if not isinstance(emitter_pump_feedback_model, PumpModel):
            raise SchemaError(
                f"{emitter_pump_feedback_model} must be of type {PumpModel}"
            )
        versioned_enum = as_enum(
            emitter_pump_feedback_model,
            EmitterPumpFeedbackModel000,
            EmitterPumpFeedbackModel000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, EmitterPumpFeedbackModel000] = {
        "00000000": EmitterPumpFeedbackModel000.ConstantDeltaT,
        "f6bde4fa": EmitterPumpFeedbackModel000.ConstantGpm,
    }

    versioned_enum_to_type_dict: Dict[EmitterPumpFeedbackModel000, str] = {
        EmitterPumpFeedbackModel000.ConstantDeltaT: "00000000",
        EmitterPumpFeedbackModel000.ConstantGpm: "f6bde4fa",
    }


class RecognizedCurrencyUnit000SchemaEnum:
    enum_name: str = "recognized.currency.unit.000"
    symbols: List[str] = [
        "00000000",
        "e57c5143",
        "f7b38fc5",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedCurrencyUnit000(StrEnum):
    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit000":
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedCurrencyUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedCurrencyUnit:
        if not RecognizedCurrencyUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedCurrencyUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )

    @classmethod
    def local_to_type(cls, recognized_currency_unit: RecognizedCurrencyUnit) -> str:
        if not isinstance(recognized_currency_unit, RecognizedCurrencyUnit):
            raise SchemaError(
                f"{recognized_currency_unit} must be of type {RecognizedCurrencyUnit}"
            )
        versioned_enum = as_enum(
            recognized_currency_unit,
            RecognizedCurrencyUnit000,
            RecognizedCurrencyUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedCurrencyUnit000] = {
        "00000000": RecognizedCurrencyUnit000.UNKNOWN,
        "e57c5143": RecognizedCurrencyUnit000.USD,
        "f7b38fc5": RecognizedCurrencyUnit000.GBP,
    }

    versioned_enum_to_type_dict: Dict[RecognizedCurrencyUnit000, str] = {
        RecognizedCurrencyUnit000.UNKNOWN: "00000000",
        RecognizedCurrencyUnit000.USD: "e57c5143",
        RecognizedCurrencyUnit000.GBP: "f7b38fc5",
    }


class RecognizedTemperatureUnit000SchemaEnum:
    enum_name: str = "recognized.temperature.unit.000"
    symbols: List[str] = [
        "00000000",
        "6f16ee63",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class RecognizedTemperatureUnit000(StrEnum):
    C = auto()
    F = auto()

    @classmethod
    def default(cls) -> "RecognizedTemperatureUnit000":
        return cls.C

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class RecognizedTemperatureUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> RecognizedTemperatureUnit:
        if not RecognizedTemperatureUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to RecognizedTemperatureUnit000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )

    @classmethod
    def local_to_type(
        cls, recognized_temperature_unit: RecognizedTemperatureUnit
    ) -> str:
        if not isinstance(recognized_temperature_unit, RecognizedTemperatureUnit):
            raise SchemaError(
                f"{recognized_temperature_unit} must be of type {RecognizedTemperatureUnit}"
            )
        versioned_enum = as_enum(
            recognized_temperature_unit,
            RecognizedTemperatureUnit000,
            RecognizedTemperatureUnit000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, RecognizedTemperatureUnit000] = {
        "00000000": RecognizedTemperatureUnit000.C,
        "6f16ee63": RecognizedTemperatureUnit000.F,
    }

    versioned_enum_to_type_dict: Dict[RecognizedTemperatureUnit000, str] = {
        RecognizedTemperatureUnit000.C: "00000000",
        RecognizedTemperatureUnit000.F: "6f16ee63",
    }


class MixingValveFeedbackModel000SchemaEnum:
    enum_name: str = "mixing.valve.feedback.model.000"
    symbols: List[str] = [
        "00000000",
        "0397c1df",
        "6a668ab8",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MixingValveFeedbackModel000(StrEnum):
    ConstantSwt = auto()
    NaiveVariableSwt = auto()
    CautiousVariableSwt = auto()

    @classmethod
    def default(cls) -> "MixingValveFeedbackModel000":
        return cls.ConstantSwt

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MixingValveFeedbackModelMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MixingValveModel:
        if not MixingValveFeedbackModel000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to MixingValveFeedbackModel000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MixingValveModel, MixingValveModel.default())

    @classmethod
    def local_to_type(cls, mixing_valve_feedback_model: MixingValveModel) -> str:
        if not isinstance(mixing_valve_feedback_model, MixingValveModel):
            raise SchemaError(
                f"{mixing_valve_feedback_model} must be of type {MixingValveModel}"
            )
        versioned_enum = as_enum(
            mixing_valve_feedback_model,
            MixingValveFeedbackModel000,
            MixingValveFeedbackModel000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MixingValveFeedbackModel000] = {
        "00000000": MixingValveFeedbackModel000.ConstantSwt,
        "0397c1df": MixingValveFeedbackModel000.NaiveVariableSwt,
        "6a668ab8": MixingValveFeedbackModel000.CautiousVariableSwt,
    }

    versioned_enum_to_type_dict: Dict[MixingValveFeedbackModel000, str] = {
        MixingValveFeedbackModel000.ConstantSwt: "00000000",
        MixingValveFeedbackModel000.NaiveVariableSwt: "0397c1df",
        MixingValveFeedbackModel000.CautiousVariableSwt: "6a668ab8",
    }


class EnergySupplyType000SchemaEnum:
    enum_name: str = "energy.supply.type.000"
    symbols: List[str] = [
        "00000000",
        "cb18f937",
        "e9dc99a6",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class EnergySupplyType000(StrEnum):
    Unknown = auto()
    StandardOffer = auto()
    RealtimeLocalLmp = auto()

    @classmethod
    def default(cls) -> "EnergySupplyType000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class EnergySupplyTypeMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> EnergySupplyType:
        if not EnergySupplyType000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to EnergySupplyType000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, EnergySupplyType, EnergySupplyType.default())

    @classmethod
    def local_to_type(cls, energy_supply_type: EnergySupplyType) -> str:
        if not isinstance(energy_supply_type, EnergySupplyType):
            raise SchemaError(
                f"{energy_supply_type} must be of type {EnergySupplyType}"
            )
        versioned_enum = as_enum(
            energy_supply_type, EnergySupplyType000, EnergySupplyType000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, EnergySupplyType000] = {
        "00000000": EnergySupplyType000.Unknown,
        "cb18f937": EnergySupplyType000.StandardOffer,
        "e9dc99a6": EnergySupplyType000.RealtimeLocalLmp,
    }

    versioned_enum_to_type_dict: Dict[EnergySupplyType000, str] = {
        EnergySupplyType000.Unknown: "00000000",
        EnergySupplyType000.StandardOffer: "cb18f937",
        EnergySupplyType000.RealtimeLocalLmp: "e9dc99a6",
    }


class FloParamsHeatpumpwithbooststore(BaseModel):
    StoreSizeGallons: int = 240  #
    MaxStoreTempF: int = 190  #
    StoreMaxPowerKw: float = 9  #
    RatedHeatpumpElectricityKw: float = 5.5  #
    MaxHeatpumpSourceWaterTempF: int = 135  #
    SystemMaxHeatOutputSwtF: int = 160  #
    SystemMaxHeatOutputDeltaTempF: int = 20  #
    SystemMaxHeatOutputGpm: float = 6  #
    EmitterMaxSafeSwtF: int = 170  #
    CirculatorPumpMaxGpm: float = 10  #
    HeatpumpTariff: DistributionTariff = DistributionTariff.VersantStorageHeatTariff  #
    HeatpumpEnergySupplyType: EnergySupplyType = EnergySupplyType.RealtimeLocalLmp  #
    BoostTariff: DistributionTariff = DistributionTariff.VersantStorageHeatTariff  #
    BoostEnergySupplyType: EnergySupplyType = EnergySupplyType.RealtimeLocalLmp  #
    StandardOfferPriceDollarsPerMwh: int = 110  #
    DistributionTariffDollarsPerMwh: int = 113  #
    AmbientTempStoreF: int = 65  #
    StorePassiveLossRatio: float = 0.001  #
    RoomTempF: int = 70  #
    AmbientPowerInKw: float = 1.25  #
    ZeroPotentialEnergyWaterTempF: int = 100  #
    EmitterPumpFeedbackModel: PumpModel = PumpModel.ConstantDeltaT  #
    MixingValveFeedbackModel: MixingValveModel = MixingValveModel.NaiveVariableSwt  #
    CautiousMixingValveTempDeltaF: int = 5  #
    Cop1TempF: int = -20  #
    Cop4TempF: int = 70  #
    CurrencyUnit: RecognizedCurrencyUnit = RecognizedCurrencyUnit.USD  #
    TempUnit: RecognizedTemperatureUnit = RecognizedTemperatureUnit.F  #
    TimezoneString: str = "US/Eastern"  #
    HomeCity: str = "MILLINOCKET_ME"  #
    StorageSteps: int = 100  #
    HouseWorstCaseTempF: int = -7  #
    SystemMaxHeatOutputKwAvg: float = 11.68  #
    K: float = -1.1  #
    IsRegulating: bool = False  #
    SliceDurationMinutes: List[int] = [60]  #
    PowerRequiredByHouseFromSystemAvgKwList: List[float] = [3.42]  #
    OutsideTempF: List[float] = [-5.1]  #
    RealtimeElectricityPrice: List[float] = [10.35]  #
    DistributionPrice: List[float] = [40.0]  #
    RegulationPrice: List[float] = [25.3]  #
    RtElecPriceUid: str  #
    WeatherUid: str  #
    DistPriceUid: Optional[str] = None
    RegPriceUid: Optional[str] = None
    StartYearUtc: int = 2020  #
    StartMonthUtc: int = 1  #
    StartDayUtc: int = 1  #
    StartHourUtc: int = 1  #
    StartMinuteUtc: int = 0  #
    StartingStoreIdx: int = 50  #
    TypeName: Literal[
        "flo.params.heatpumpwithbooststore"
    ] = "flo.params.heatpumpwithbooststore"
    Version: str = "000"

    @validator("HeatpumpTariff")
    def _validator_heatpump_tariff(cls, v: DistributionTariff) -> DistributionTariff:
        return as_enum(v, DistributionTariff, DistributionTariff.Unknown)

    @validator("HeatpumpEnergySupplyType")
    def _validator_heatpump_energy_supply_type(
        cls, v: EnergySupplyType
    ) -> EnergySupplyType:
        return as_enum(v, EnergySupplyType, EnergySupplyType.Unknown)

    @validator("BoostTariff")
    def _validator_boost_tariff(cls, v: DistributionTariff) -> DistributionTariff:
        return as_enum(v, DistributionTariff, DistributionTariff.Unknown)

    @validator("BoostEnergySupplyType")
    def _validator_boost_energy_supply_type(
        cls, v: EnergySupplyType
    ) -> EnergySupplyType:
        return as_enum(v, EnergySupplyType, EnergySupplyType.Unknown)

    @validator("EmitterPumpFeedbackModel")
    def _validator_emitter_pump_feedback_model(
        cls, v: EmitterPumpFeedbackModel
    ) -> EmitterPumpFeedbackModel:
        return as_enum(v, PumpModel, PumpModel.ConstantDeltaT)

    @validator("MixingValveFeedbackModel")
    def _validator_mixing_valve_feedback_model(
        cls, v: MixingValveFeedbackModel
    ) -> MixingValveFeedbackModel:
        return as_enum(v, MixingValveModel, MixingValveModel.ConstantSwt)

    @validator("CurrencyUnit")
    def _validator_currency_unit(
        cls, v: RecognizedCurrencyUnit
    ) -> RecognizedCurrencyUnit:
        return as_enum(v, RecognizedCurrencyUnit, RecognizedCurrencyUnit.UNKNOWN)

    @validator("TempUnit")
    def _validator_temp_unit(
        cls, v: RecognizedTemperatureUnit
    ) -> RecognizedTemperatureUnit:
        return as_enum(v, RecognizedTemperatureUnit, RecognizedTemperatureUnit.C)

    _validator_rt_elec_price_uid = predicate_validator(
        "RtElecPriceUid", property_format.is_uuid_canonical_textual
    )

    _validator_weather_uid = predicate_validator(
        "WeatherUid", property_format.is_uuid_canonical_textual
    )

    @validator("DistPriceUid")
    def _validator_dist_price_uid(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if not property_format.is_uuid_canonical_textual(v):
            raise ValueError(f"DistPriceUid {v} must have UuidCanonicalTextual")
        return v

    @validator("RegPriceUid")
    def _validator_reg_price_uid(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if not property_format.is_uuid_canonical_textual(v):
            raise ValueError(f"RegPriceUid {v} must have UuidCanonicalTextual")
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["HeatpumpTariff"]
        HeatpumpTariff = as_enum(
            self.HeatpumpTariff, DistributionTariff, DistributionTariff.default()
        )
        d["HeatpumpTariffGtEnumSymbol"] = DistributionTariffMap.local_to_type(
            HeatpumpTariff
        )
        del d["HeatpumpEnergySupplyType"]
        HeatpumpEnergySupplyType = as_enum(
            self.HeatpumpEnergySupplyType, EnergySupplyType, EnergySupplyType.default()
        )
        d["HeatpumpEnergySupplyTypeGtEnumSymbol"] = EnergySupplyTypeMap.local_to_type(
            HeatpumpEnergySupplyType
        )
        del d["BoostTariff"]
        BoostTariff = as_enum(
            self.BoostTariff, DistributionTariff, DistributionTariff.default()
        )
        d["BoostTariffGtEnumSymbol"] = DistributionTariffMap.local_to_type(BoostTariff)
        del d["BoostEnergySupplyType"]
        BoostEnergySupplyType = as_enum(
            self.BoostEnergySupplyType, EnergySupplyType, EnergySupplyType.default()
        )
        d["BoostEnergySupplyTypeGtEnumSymbol"] = EnergySupplyTypeMap.local_to_type(
            BoostEnergySupplyType
        )
        del d["EmitterPumpFeedbackModel"]
        EmitterPumpFeedbackModel = as_enum(
            self.EmitterPumpFeedbackModel,
            PumpModel,
            PumpModel.default(),
        )
        d[
            "EmitterPumpFeedbackModelGtEnumSymbol"
        ] = EmitterPumpFeedbackModelMap.local_to_type(EmitterPumpFeedbackModel)
        del d["MixingValveFeedbackModel"]
        MixingValveFeedbackModel = as_enum(
            self.MixingValveFeedbackModel,
            MixingValveModel,
            MixingValveModel.default(),
        )
        d[
            "MixingValveFeedbackModelGtEnumSymbol"
        ] = MixingValveFeedbackModelMap.local_to_type(MixingValveFeedbackModel)
        del d["CurrencyUnit"]
        CurrencyUnit = as_enum(
            self.CurrencyUnit, RecognizedCurrencyUnit, RecognizedCurrencyUnit.default()
        )
        d["CurrencyUnitGtEnumSymbol"] = RecognizedCurrencyUnitMap.local_to_type(
            CurrencyUnit
        )
        del d["TempUnit"]
        TempUnit = as_enum(
            self.TempUnit,
            RecognizedTemperatureUnit,
            RecognizedTemperatureUnit.default(),
        )
        d["TempUnitGtEnumSymbol"] = RecognizedTemperatureUnitMap.local_to_type(TempUnit)
        if d["DistPriceUid"] is None:
            del d["DistPriceUid"]
        if d["RegPriceUid"] is None:
            del d["RegPriceUid"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class FloParamsHeatpumpwithbooststore_Maker:
    type_name = "flo.params.heatpumpwithbooststore"
    version = "000"

    def __init__(
        self,
        store_size_gallons: int,
        max_store_temp_f: int,
        store_max_power_kw: float,
        rated_heatpump_electricity_kw: float,
        max_heatpump_source_water_temp_f: int,
        system_max_heat_output_swt_f: int,
        system_max_heat_output_delta_temp_f: int,
        system_max_heat_output_gpm: float,
        emitter_max_safe_swt_f: int,
        circulator_pump_max_gpm: float,
        heatpump_tariff: DistributionTariff,
        heatpump_energy_supply_type: EnergySupplyType,
        boost_tariff: DistributionTariff,
        boost_energy_supply_type: EnergySupplyType,
        standard_offer_price_dollars_per_mwh: int,
        distribution_tariff_dollars_per_mwh: int,
        ambient_temp_store_f: int,
        store_passive_loss_ratio: float,
        room_temp_f: int,
        ambient_power_in_kw: float,
        zero_potential_energy_water_temp_f: int,
        emitter_pump_feedback_model: PumpModel,
        mixing_valve_feedback_model: MixingValveModel,
        cautious_mixing_valve_temp_delta_f: int,
        cop1_temp_f: int,
        cop4_temp_f: int,
        currency_unit: RecognizedCurrencyUnit,
        temp_unit: RecognizedTemperatureUnit,
        timezone_string: str,
        home_city: str,
        storage_steps: int,
        house_worst_case_temp_f: int,
        system_max_heat_output_kw_avg: float,
        k: float,
        is_regulating: bool,
        slice_duration_minutes: List[int],
        power_required_by_house_from_system_avg_kw_list: List[float],
        outside_temp_f: List[float],
        realtime_electricity_price: List[float],
        distribution_price: List[float],
        regulation_price: List[float],
        rt_elec_price_uid: str,
        weather_uid: str,
        dist_price_uid: Optional[str],
        reg_price_uid: Optional[str],
        start_year_utc: int,
        start_month_utc: int,
        start_day_utc: int,
        start_hour_utc: int,
        start_minute_utc: int,
        starting_store_idx: int,
    ):
        self.tuple = FloParamsHeatpumpwithbooststore(
            StoreSizeGallons=store_size_gallons,
            MaxStoreTempF=max_store_temp_f,
            StoreMaxPowerKw=store_max_power_kw,
            RatedHeatpumpElectricityKw=rated_heatpump_electricity_kw,
            MaxHeatpumpSourceWaterTempF=max_heatpump_source_water_temp_f,
            SystemMaxHeatOutputSwtF=system_max_heat_output_swt_f,
            SystemMaxHeatOutputDeltaTempF=system_max_heat_output_delta_temp_f,
            SystemMaxHeatOutputGpm=system_max_heat_output_gpm,
            EmitterMaxSafeSwtF=emitter_max_safe_swt_f,
            CirculatorPumpMaxGpm=circulator_pump_max_gpm,
            HeatpumpTariff=heatpump_tariff,
            HeatpumpEnergySupplyType=heatpump_energy_supply_type,
            BoostTariff=boost_tariff,
            BoostEnergySupplyType=boost_energy_supply_type,
            StandardOfferPriceDollarsPerMwh=standard_offer_price_dollars_per_mwh,
            DistributionTariffDollarsPerMwh=distribution_tariff_dollars_per_mwh,
            AmbientTempStoreF=ambient_temp_store_f,
            StorePassiveLossRatio=store_passive_loss_ratio,
            RoomTempF=room_temp_f,
            AmbientPowerInKw=ambient_power_in_kw,
            ZeroPotentialEnergyWaterTempF=zero_potential_energy_water_temp_f,
            EmitterPumpFeedbackModel=emitter_pump_feedback_model,
            MixingValveFeedbackModel=mixing_valve_feedback_model,
            CautiousMixingValveTempDeltaF=cautious_mixing_valve_temp_delta_f,
            Cop1TempF=cop1_temp_f,
            Cop4TempF=cop4_temp_f,
            CurrencyUnit=currency_unit,
            TempUnit=temp_unit,
            TimezoneString=timezone_string,
            HomeCity=home_city,
            StorageSteps=storage_steps,
            HouseWorstCaseTempF=house_worst_case_temp_f,
            SystemMaxHeatOutputKwAvg=system_max_heat_output_kw_avg,
            K=k,
            IsRegulating=is_regulating,
            SliceDurationMinutes=slice_duration_minutes,
            PowerRequiredByHouseFromSystemAvgKwList=power_required_by_house_from_system_avg_kw_list,
            OutsideTempF=outside_temp_f,
            RealtimeElectricityPrice=realtime_electricity_price,
            DistributionPrice=distribution_price,
            RegulationPrice=regulation_price,
            RtElecPriceUid=rt_elec_price_uid,
            WeatherUid=weather_uid,
            DistPriceUid=dist_price_uid,
            RegPriceUid=reg_price_uid,
            StartYearUtc=start_year_utc,
            StartMonthUtc=start_month_utc,
            StartDayUtc=start_day_utc,
            StartHourUtc=start_hour_utc,
            StartMinuteUtc=start_minute_utc,
            StartingStoreIdx=starting_store_idx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: FloParamsHeatpumpwithbooststore) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> FloParamsHeatpumpwithbooststore:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> FloParamsHeatpumpwithbooststore:
        d2 = dict(d)
        if "StoreSizeGallons" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StoreSizeGallons")
        if "MaxStoreTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MaxStoreTempF")
        if "StoreMaxPowerKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StoreMaxPowerKw")
        if "RatedHeatpumpElectricityKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RatedHeatpumpElectricityKw")
        if "MaxHeatpumpSourceWaterTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MaxHeatpumpSourceWaterTempF")
        if "SystemMaxHeatOutputSwtF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SystemMaxHeatOutputSwtF")
        if "SystemMaxHeatOutputDeltaTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SystemMaxHeatOutputDeltaTempF")
        if "SystemMaxHeatOutputGpm" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SystemMaxHeatOutputGpm")
        if "EmitterMaxSafeSwtF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing EmitterMaxSafeSwtF")
        if "CirculatorPumpMaxGpm" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CirculatorPumpMaxGpm")
        if "HeatpumpTariffGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HeatpumpTariffGtEnumSymbol")
        if d2["HeatpumpTariffGtEnumSymbol"] in DistributionTariff000SchemaEnum.symbols:
            d2["HeatpumpTariff"] = DistributionTariffMap.type_to_local(
                d2["HeatpumpTariffGtEnumSymbol"]
            )
        else:
            d2["HeatpumpTariff"] = DistributionTariff.default()
        if "HeatpumpEnergySupplyTypeGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HeatpumpEnergySupplyTypeGtEnumSymbol")
        if (
            d2["HeatpumpEnergySupplyTypeGtEnumSymbol"]
            in EnergySupplyType000SchemaEnum.symbols
        ):
            d2["HeatpumpEnergySupplyType"] = EnergySupplyTypeMap.type_to_local(
                d2["HeatpumpEnergySupplyTypeGtEnumSymbol"]
            )
        else:
            d2["HeatpumpEnergySupplyType"] = EnergySupplyType.default()
        if "BoostTariffGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BoostTariffGtEnumSymbol")
        if d2["BoostTariffGtEnumSymbol"] in DistributionTariff000SchemaEnum.symbols:
            d2["BoostTariff"] = DistributionTariffMap.type_to_local(
                d2["BoostTariffGtEnumSymbol"]
            )
        else:
            d2["BoostTariff"] = DistributionTariff.default()
        if "BoostEnergySupplyTypeGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BoostEnergySupplyTypeGtEnumSymbol")
        if (
            d2["BoostEnergySupplyTypeGtEnumSymbol"]
            in EnergySupplyType000SchemaEnum.symbols
        ):
            d2["BoostEnergySupplyType"] = EnergySupplyTypeMap.type_to_local(
                d2["BoostEnergySupplyTypeGtEnumSymbol"]
            )
        else:
            d2["BoostEnergySupplyType"] = EnergySupplyType.default()
        if "StandardOfferPriceDollarsPerMwh" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StandardOfferPriceDollarsPerMwh")
        if "DistributionTariffDollarsPerMwh" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DistributionTariffDollarsPerMwh")
        if "AmbientTempStoreF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AmbientTempStoreF")
        if "StorePassiveLossRatio" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StorePassiveLossRatio")
        if "RoomTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RoomTempF")
        if "AmbientPowerInKw" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AmbientPowerInKw")
        if "ZeroPotentialEnergyWaterTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ZeroPotentialEnergyWaterTempF")
        if "EmitterPumpFeedbackModelGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing EmitterPumpFeedbackModelGtEnumSymbol")
        if (
            d2["EmitterPumpFeedbackModelGtEnumSymbol"]
            in EmitterPumpFeedbackModel000SchemaEnum.symbols
        ):
            d2["EmitterPumpFeedbackModel"] = EmitterPumpFeedbackModelMap.type_to_local(
                d2["EmitterPumpFeedbackModelGtEnumSymbol"]
            )
        else:
            d2["EmitterPumpFeedbackModel"] = PumpModel.default()
        if "MixingValveFeedbackModelGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MixingValveFeedbackModelGtEnumSymbol")
        if (
            d2["MixingValveFeedbackModelGtEnumSymbol"]
            in MixingValveFeedbackModel000SchemaEnum.symbols
        ):
            d2["MixingValveFeedbackModel"] = MixingValveFeedbackModelMap.type_to_local(
                d2["MixingValveFeedbackModelGtEnumSymbol"]
            )
        else:
            d2["MixingValveFeedbackModel"] = MixingValveModel.default()
        if "CautiousMixingValveTempDeltaF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CautiousMixingValveTempDeltaF")
        if "Cop1TempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Cop1TempF")
        if "Cop4TempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Cop4TempF")
        if "CurrencyUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing CurrencyUnitGtEnumSymbol")
        if (
            d2["CurrencyUnitGtEnumSymbol"]
            in RecognizedCurrencyUnit000SchemaEnum.symbols
        ):
            d2["CurrencyUnit"] = RecognizedCurrencyUnitMap.type_to_local(
                d2["CurrencyUnitGtEnumSymbol"]
            )
        else:
            d2["CurrencyUnit"] = RecognizedCurrencyUnit.default()
        if "TempUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TempUnitGtEnumSymbol")
        if d2["TempUnitGtEnumSymbol"] in RecognizedTemperatureUnit000SchemaEnum.symbols:
            d2["TempUnit"] = RecognizedTemperatureUnitMap.type_to_local(
                d2["TempUnitGtEnumSymbol"]
            )
        else:
            d2["TempUnit"] = RecognizedTemperatureUnit.default()
        if "TimezoneString" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TimezoneString")
        if "HomeCity" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HomeCity")
        if "StorageSteps" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StorageSteps")
        if "HouseWorstCaseTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HouseWorstCaseTempF")
        if "SystemMaxHeatOutputKwAvg" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SystemMaxHeatOutputKwAvg")
        if "K" not in d2.keys():
            raise SchemaError(f"dict {d2} missing K")
        if "IsRegulating" not in d2.keys():
            raise SchemaError(f"dict {d2} missing IsRegulating")
        if "SliceDurationMinutes" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SliceDurationMinutes")
        if "PowerRequiredByHouseFromSystemAvgKwList" not in d2.keys():
            raise SchemaError(
                f"dict {d2} missing PowerRequiredByHouseFromSystemAvgKwList"
            )
        if "OutsideTempF" not in d2.keys():
            raise SchemaError(f"dict {d2} missing OutsideTempF")
        if "RealtimeElectricityPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RealtimeElectricityPrice")
        if "DistributionPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing DistributionPrice")
        if "RegulationPrice" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RegulationPrice")
        if "RtElecPriceUid" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RtElecPriceUid")
        if "WeatherUid" not in d2.keys():
            raise SchemaError(f"dict {d2} missing WeatherUid")
        if "DistPriceUid" not in d2.keys():
            d2["DistPriceUid"] = None
        if "RegPriceUid" not in d2.keys():
            d2["RegPriceUid"] = None
        if "StartYearUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartYearUtc")
        if "StartMonthUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartMonthUtc")
        if "StartDayUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartDayUtc")
        if "StartHourUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartHourUtc")
        if "StartMinuteUtc" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartMinuteUtc")
        if "StartingStoreIdx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartingStoreIdx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return FloParamsHeatpumpwithbooststore(
            StoreSizeGallons=d2["StoreSizeGallons"],
            MaxStoreTempF=d2["MaxStoreTempF"],
            StoreMaxPowerKw=d2["StoreMaxPowerKw"],
            RatedHeatpumpElectricityKw=d2["RatedHeatpumpElectricityKw"],
            MaxHeatpumpSourceWaterTempF=d2["MaxHeatpumpSourceWaterTempF"],
            SystemMaxHeatOutputSwtF=d2["SystemMaxHeatOutputSwtF"],
            SystemMaxHeatOutputDeltaTempF=d2["SystemMaxHeatOutputDeltaTempF"],
            SystemMaxHeatOutputGpm=d2["SystemMaxHeatOutputGpm"],
            EmitterMaxSafeSwtF=d2["EmitterMaxSafeSwtF"],
            CirculatorPumpMaxGpm=d2["CirculatorPumpMaxGpm"],
            HeatpumpTariff=d2["HeatpumpTariff"],
            HeatpumpEnergySupplyType=d2["HeatpumpEnergySupplyType"],
            BoostTariff=d2["BoostTariff"],
            BoostEnergySupplyType=d2["BoostEnergySupplyType"],
            StandardOfferPriceDollarsPerMwh=d2["StandardOfferPriceDollarsPerMwh"],
            DistributionTariffDollarsPerMwh=d2["DistributionTariffDollarsPerMwh"],
            AmbientTempStoreF=d2["AmbientTempStoreF"],
            StorePassiveLossRatio=d2["StorePassiveLossRatio"],
            RoomTempF=d2["RoomTempF"],
            AmbientPowerInKw=d2["AmbientPowerInKw"],
            ZeroPotentialEnergyWaterTempF=d2["ZeroPotentialEnergyWaterTempF"],
            EmitterPumpFeedbackModel=d2["EmitterPumpFeedbackModel"],
            MixingValveFeedbackModel=d2["MixingValveFeedbackModel"],
            CautiousMixingValveTempDeltaF=d2["CautiousMixingValveTempDeltaF"],
            Cop1TempF=d2["Cop1TempF"],
            Cop4TempF=d2["Cop4TempF"],
            CurrencyUnit=d2["CurrencyUnit"],
            TempUnit=d2["TempUnit"],
            TimezoneString=d2["TimezoneString"],
            HomeCity=d2["HomeCity"],
            StorageSteps=d2["StorageSteps"],
            HouseWorstCaseTempF=d2["HouseWorstCaseTempF"],
            SystemMaxHeatOutputKwAvg=d2["SystemMaxHeatOutputKwAvg"],
            K=d2["K"],
            IsRegulating=d2["IsRegulating"],
            SliceDurationMinutes=d2["SliceDurationMinutes"],
            PowerRequiredByHouseFromSystemAvgKwList=d2[
                "PowerRequiredByHouseFromSystemAvgKwList"
            ],
            OutsideTempF=d2["OutsideTempF"],
            RealtimeElectricityPrice=d2["RealtimeElectricityPrice"],
            DistributionPrice=d2["DistributionPrice"],
            RegulationPrice=d2["RegulationPrice"],
            RtElecPriceUid=d2["RtElecPriceUid"],
            WeatherUid=d2["WeatherUid"],
            DistPriceUid=d2["DistPriceUid"],
            RegPriceUid=d2["RegPriceUid"],
            StartYearUtc=d2["StartYearUtc"],
            StartMonthUtc=d2["StartMonthUtc"],
            StartDayUtc=d2["StartDayUtc"],
            StartHourUtc=d2["StartHourUtc"],
            StartMinuteUtc=d2["StartMinuteUtc"],
            StartingStoreIdx=d2["StartingStoreIdx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class UniverseType(StrEnum):
    """
    Allows for multiple GridWorks, in particular for development and shared simulations. [More Info](https://gridworks.readthedocs.io/en/latest/universe.html).

    Choices and descriptions:

      * Dev: Simulation running on a single computer.
      * Hybrid: Anything goes.
      * Production: Money at stake.
    """

    Dev = auto()
    Hybrid = auto()
    Production = auto()

    @classmethod
    def default(cls) -> "UniverseType":
        """
        Returns default value Dev
        """
        return cls.Dev

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MixingValveFeedbackModel(StrEnum):
    """
    Control mechanism for a mixing valve, used by Spaceheat SCADAs

    Choices and descriptions:

      * ConstantSwt: Constant Source Water Temp
      * NaiveVariableSwt: Variable Source Water Temp, naive assumptions about distribution system capabilities
      * CautiousVariableSwt: Variable Source Water Temp, conservative assumptions about distribution system capabilities
    """

    ConstantSwt = auto()
    NaiveVariableSwt = auto()
    CautiousVariableSwt = auto()

    @classmethod
    def default(cls) -> "MixingValveFeedbackModel":
        """
        Returns default value ConstantSwt
        """
        return cls.ConstantSwt

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



"""Gridworks AtomicTNode"""
import gwatn.config as config
import gwatn.enums as enums
from gwatn.atn_actor_base import AtnActorBase
from gwatn.dispatch_contract import DispatchContract
from gwatn.two_channel_actor_base import TwoChannelActorBase


__all__ = [
    "config",
    "enums",
    "AtnActorBase",
    "TwoChannelActorBase",
]



import logging
import subprocess

import gridworks.algo_utils as algo_utils
import gridworks.dev_utils.algo_setup as algo_setup
import gridworks.errors as errors
import requests
from algosdk import encoding
from algosdk.future.transaction import *
from algosdk.v2client.algod import AlgodClient
from gridworks.algo_utils import BasicAccount
from gridworks.algo_utils import MultisigAccount
from gridworks.utils import RestfulResponse

import gwatn.config as config
from gwatn.dispatch_contract import DispatchContract

# Sent by homeowner
from gwatn.types import BaseGNodeGt_Maker
from gwatn.types import BasegnodeScadaCreate_Maker
from gwatn.types import InitialTadeedAlgoOptin_Maker
from gwatn.types import SlaEnter_Maker
from gwatn.types import TerminalassetCertifyHack_Maker


LOGGER = logging.getLogger(__name__)


class DevTaOwner:
    def __init__(
        self,
        settings: config.TaOwnerSettings,
    ):
        LOGGER.info(f"Initializing TaOwner for {settings.initial_ta_alias}")

        self.settings = settings
        self.scada_sk_by_ta_alias = {
            "d1.isone.ver.keene.holly.ta": "7RaOTVzqiL5NHarfkp7FpR1rjmn3T4Nibd+017Uhd9GpOVe2and91D46zJxO1wycU02TXxVu18cmslBEKgb8IA==",
            "d1.isone.ver.keene.juniper.ta": "K1tymo7Y9Lr2ACt67JGgUYAgeZA+Ufr9vICxn4643TI8Cxq6Mk/fdHwqOBByarqvGT80Q4GZTeyuQBWwwre54w==",
            "d1.isone.ver.keene.violet.ta": "sWThfGj1CeWD1MU1ZRkzQjBrP5F3G5Ij1XewTqp8FCr7bDmlfuY5G1GQju6rBNHVs5DKshtEaWyBgvXqV7aeNA==",
            "d1.isone.ver.keene.lettuce.ta": "GHY1eOFzSQsWM2BlhLGlNJc5JhAFN9yP0suUZ9usCnuisN9OiU/6TyHcuZczZDLFBi6LZCEzWActfaiMmfaKKw==",
        }

        self.scada_addr_by_ta_alias = {
            "d1.isone.ver.keene.holly.ta": "VE4VPNTKO565IPR2ZSOE5VYMTRJU3E27CVXNPRZGWJIEIKQG7QQKFVXDYE",
            "d1.isone.ver.keene.juniper.ta": "HQFRVORSJ7PXI7BKHAIHE2V2V4MT6NCDQGMU33FOIAK3BQVXXHRWJY6VWI",
            "d1.isone.ver.keene.violet.ta": "7NWDTJL64Y4RWUMQR3XKWBGR2WZZBSVSDNCGS3EBQL26UV5WTY2CUK6OJI",
            "d1.isone.ver.keene.lettuce.ta": "UKYN6TUJJ75E6IO4XGLTGZBSYUDC5C3EEEZVQBZNPWUIZGPWRIVQSOK6QA",
        }
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.sp = self.client.suggested_params()
        self.acct: BasicAccount = BasicAccount(
            private_key=self.settings.sk.get_secret_value()
        )

        algos = (
            self.settings.public.ta_deed_consideration_algos
            + DispatchContract.min_balance
            + 11
        )
        algo_setup.dev_fund_to_min(self.acct.addr, algos)
        LOGGER.info(
            f"HollyHomeowner acct {self.acct.addr_short_hand} balance: ~{algo_utils.algos(self.acct.addr)} Algos"
        )

        self.validator_multi = MultisigAccount(
            version=1,
            threshold=2,
            addresses=[
                self.settings.public.gnf_admin_addr,
                self.settings.ta_validator_addr,
            ],
        )
        if self.settings.initial_ta_alias == "d1.isone.ver.keene.holly.ta":
            ta_daemon_sk = config.TaDaemonSettings().sk.get_secret_value()
            self.ta_daemon_acct = BasicAccount(ta_daemon_sk)
        else:
            self.ta_daemon_acct: BasicAccount = BasicAccount()
            self.settings.ta_daemon_addr = self.ta_daemon_acct.addr

        self.ta_daemon_api_root = (
            f"{self.settings.ta_daemon_api_fqdn}:{self.settings.ta_daemon_api_port}"
        )
        LOGGER.info(f"ta_owner_addr = {self.acct.addr}")
        LOGGER.info(f"ta_daemon_addr = {self.ta_daemon_acct.addr}")

    def start(self):
        LOGGER.info(
            f"Starting {self.short_alias}-daemon on port {self.settings.ta_daemon_api_port}"
        )
        self.pr: subprocess.Popen = self.start_ta_daemon()

    def stop(self):
        # self.pr.terminate()
        cmd = f"docker stop {self.short_alias}-daemon"
        subprocess.run(cmd.split())
        cmd = f"docker rm {self.short_alias}-daemon"
        subprocess.run(cmd.split())
        cmd = f"docker stop {self.short_alias}-scada"
        subprocess.run(cmd.split())
        cmd = f"docker rm {self.short_alias}-scada"
        subprocess.run(cmd.split())

    def start_scada_actor(self, scada_cert_idx: int) -> subprocess.Popen:
        LOGGER.info(f"Starting ScadaActor with SCADA CERT {scada_cert_idx}")
        ta_alias = self.settings.initial_ta_alias
        scada_alias = ta_alias + ".scada"
        scada_sk = self.scada_sk_by_ta_alias[ta_alias]
        cmd = f"docker run  -e SCADA_CERT_IDX={scada_cert_idx} -e SCADA_G_NODE_ALIAS={scada_alias} -e SCADA_SK={scada_sk} --name {self.short_alias}-scada jessmillar/scada:latest"
        pr = subprocess.Popen(
            cmd.split(),
        )
        return pr

    def start_ta_daemon(self) -> subprocess.Popen:
        LOGGER.info("Starting TaDaemon")
        port = self.settings.ta_daemon_api_port
        cmd = f"docker run  -e TAD_SK={self.ta_daemon_acct.sk} -e TAD_TA_OWNER_ADDR={self.acct.addr} -p {port}:8000 --name {self.short_alias}-daemon jessmillar/python-ta-daemon:chaos__2e38992__20221231"
        pr = subprocess.Popen(
            cmd.split(),
        )

        daemon_api_root = (
            f"{self.settings.ta_daemon_api_fqdn}:{self.settings.ta_daemon_api_port}"
        )
        api_endpoint = f"{daemon_api_root}/"
        daemon_up: bool = False
        while not daemon_up:
            try:
                daemon_up = True
                requests.get(url=api_endpoint)
            except:
                daemon_up = False
        return pr

    def __repr__(self) -> str:
        return self.short_alias

    ##########################
    # Messages Sent
    ##########################

    def create_scada_g_node(self) -> RestfulResponse:
        txn = PaymentTxn(
            sender=self.acct.addr,
            sp=self.sp,
            receiver=self.scada_addr_by_ta_alias[self.settings.initial_ta_alias],
            amt=DispatchContract.min_balance + 10_000_000,
        )
        signed_txn = txn.sign(self.acct.sk)
        ta_alias = self.settings.initial_ta_alias
        try:
            self.client.send_transaction(signed_txn)
        except Exception as e:
            note = f"Algorand Failure sending transaction: {e}"
            raise errors.AlgoError(note)
        payload = BasegnodeScadaCreate_Maker(
            ta_alias=ta_alias,
            scada_addr=self.scada_addr_by_ta_alias[ta_alias],
            ta_daemon_addr=self.settings.ta_daemon_addr,
            g_node_registry_addr=self.settings.public.gnr_addr,
            signed_proof=encoding.msgpack_encode(signed_txn),
        ).tuple
        api_endpoint = f"{self.settings.public.gnf_api_root}/basegnode-scada-create/"
        r = requests.post(url=api_endpoint, json=payload.as_dict())
        if r.status_code > 200:
            if r.status_code == 422:
                note = f"Error entering SLA: " + r.json()["detail"]
            else:
                note = r.reason
            rr = RestfulResponse(Note=note, HttpStatusCode=422)
            return rr
        rr = RestfulResponse(**r.json())
        if rr.PayloadTypeName != "base.g.node.gt":
            return RestfulResponse(
                Note=f"expected base.g.node.gt as PayloadTypeName, got {rr.PayloadTypeName}",
                HttpStatusCode=422,
            )
        scada_gt = BaseGNodeGt_Maker.dict_to_tuple(rr.PayloadAsDict)
        if scada_gt.ScadaCertId is None:
            return RestfulResponse(
                Note=f"expected ScadaCertId in returned BaseGNodeGt, got None",
                HttpStatusCode=422,
            )
        LOGGER.info(
            f"Successful request for Scada. ScadaCertId is {scada_gt.ScadaCertId}"
        )
        self.start_scada_actor(scada_cert_idx=scada_gt.ScadaCertId)
        # LOGGER.info(f"Started Scada actor for {scada_gt.Alias}")

        return rr

    def enter_sla(self) -> RestfulResponse:
        ta_alias = self.settings.initial_ta_alias
        payload = SlaEnter_Maker(terminal_asset_alias=ta_alias).tuple
        api_endpoint = f"{self.ta_daemon_api_root}/sla-enter/"
        r = requests.post(url=api_endpoint, json=payload.as_dict())
        if r.status_code > 200:
            if r.status_code == 422:
                note = f"Error entering SLA: " + r.json()["detail"]
            else:
                note = r.reason
            rr = RestfulResponse(Note=note, HttpStatusCode=422)
            return rr
        return RestfulResponse(**r.json())

    def request_ta_certification(self) -> RestfulResponse:
        ta_alias = self.settings.initial_ta_alias
        payload = TerminalassetCertifyHack_Maker(
            terminal_asset_alias=ta_alias,
            ta_daemon_api_fqdn=self.settings.ta_daemon_api_fqdn,
            ta_daemon_api_port=self.settings.ta_daemon_api_port,
            ta_daemon_addr=self.settings.ta_daemon_addr,
        ).tuple

        LOGGER.info(f"Requesting certification for {ta_alias}")
        api_endpoint = (
            f"{self.settings.ta_validator_api_root}/terminalasset-certification/"
        )
        try:
            r = requests.post(url=api_endpoint, json=payload.as_dict())
        except:
            return RestfulResponse(
                Note=f"Validator address {api_endpoint} does not exist!",
                HttpStatusCode=422,
            )
        if r.status_code > 200:
            if r.status_code == 422:
                note = (
                    f"Error posting terminalasset-certification to validator: "
                    + r.json()["detail"]
                )
            else:
                note = r.reason
            rr = RestfulResponse(Note=note, HttpStatusCode=422)
            return rr
        rr1 = RestfulResponse(**r.json())
        rr2 = self.post_initial_tadeed_algo_optin()
        if rr2.HttpStatusCode > 200:
            return rr2
        return rr1

    def post_initial_tadeed_algo_optin(self) -> RestfulResponse:
        """
         - Sends 50 algos to TaDaemon acct
         - Sends InitialTadeedAlgoOptin to TaDaemon, with signed
         funding txn for proof of identity.

        Returns:
            RestfulResponse
        """
        LOGGER.info("Funding TaDaemon")
        required_algos = config.Public().ta_deed_consideration_algos
        txn = PaymentTxn(
            sender=self.acct.addr,
            receiver=self.settings.ta_daemon_addr,
            amt=required_algos * 10**6,
            sp=self.sp,
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except Exception as e:
            note = f"Algorand Failure sending transaction: {e}"
            raise errors.AlgoError(note)
        algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())
        payload = InitialTadeedAlgoOptin_Maker(
            terminal_asset_alias=self.settings.initial_ta_alias,
            ta_owner_addr=self.acct.addr,
            validator_addr=self.settings.ta_validator_addr,
            signed_initial_daemon_funding_txn=encoding.msgpack_encode(signed_txn),
        ).tuple
        api_endpoint = f"{self.ta_daemon_api_root}/initial-tadeed-algo-optin/"
        r = requests.post(url=api_endpoint, json=payload.as_dict())
        if r.status_code > 200:
            if r.status_code == 422:
                note = "Issue with InitialTadeedAlgoOptin" + r.json()["detail"]
            else:
                note = r.reason
            rr = RestfulResponse(Note=note, HttpStatusCode=422)
            return rr
        return RestfulResponse(**r.json())

    ##########################
    # dev methods
    ########################

    @property
    def short_alias(self) -> str:
        alias = self.settings.initial_ta_alias
        words = alias.split(".")
        return ".".join(words[-2:-1])



import logging
import time
from typing import Optional

import dotenv
import gridworks.algo_utils as algo_utils
from algosdk.v2client.algod import AlgodClient
from beaker.client import ApplicationClient
from gridworks.algo_utils import BasicAccount
from gridworks.enums import GNodeRole

import gwatn.config as config
from gwatn import DispatchContract
from gwatn.atn_actor_base import AtnActorBase
from gwatn.types import LatestPrice
from gwatn.types import SimTimestep


LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class SimpleAtnActor(AtnActorBase):
    """Simple implementation of an AtnActor, for testing purposes"""

    def __init__(
        self,
        settings: config.AtnSettings = config.AtnSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings=settings)
        LOGGER.info("Simple Atn Initialized")

    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        pass

    def new_timestep(self, payload: SimTimestep) -> None:
        """Set to work with a timestep per minute"""

        # sends a hb to Scada every minute for DispatchContract
        self.hb_to_scada()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        pass



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class Role(StrEnum):
    """
    Determines the functional role of a SpaceHeat node supervised by a SCADA

    Choices and descriptions:

      * PrimaryScada:
      * HomeAlone:
      * PrimaryMeter:
      * BoostElement:
      * BooleanActuator:
      * Atn:
      * TankWaterTempSensor:
      * PipeTempSensor:
      * RoomTempSensor:
      * OutdoorTempSensor:
      * Heatpump:
      * MultipurposeSensor:
      * CurrentTransformer:
      * RadiatorFan:
      * PipeFlowMeter:
      * BaseboardRadiator:
      * CirculatorPump:
      * DedicatedThermalStore:
      * HeatedSpace:
      * HydronicPipe:
      * Outdoors:
    """

    PrimaryScada = auto()
    HomeAlone = auto()
    Heatpump = auto()
    MultipurposeSensor = auto()
    CurrentTransformer = auto()
    RadiatorFan = auto()
    PipeFlowMeter = auto()
    BaseboardRadiator = auto()
    CirculatorPump = auto()
    DedicatedThermalStore = auto()
    HeatedSpace = auto()
    HydronicPipe = auto()
    PrimaryMeter = auto()
    Outdoors = auto()
    BoostElement = auto()
    BooleanActuator = auto()
    Atn = auto()
    TankWaterTempSensor = auto()
    PipeTempSensor = auto()
    RoomTempSensor = auto()
    OutdoorTempSensor = auto()

    @classmethod
    def default(cls) -> "Role":
        """
        Returns default value HeatedSpace
        """
        return cls.HeatedSpace

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class TelemetryName(StrEnum):
    """
    Specifies the name of sensed data reported by SCADA

    Choices and descriptions:

      * Unknown:
      * PowerW:
      * RelayState:
      * WaterTempCTimes1000:
      * WaterTempFTimes1000:
      * WaterFlowGpmTimes100:
      * CurrentRmsMicroAmps:
      * GallonsPerMinuteTimes10:
      * CurrentRmsMilliAmps:
      * VoltageRmsMilliVolts:
      * PhaseAngleDegreesTimes10:
      * WattHours:
    """

    Unknown = auto()
    PowerW = auto()
    PhaseAngleDegreesTimes10 = auto()
    WattHours = auto()
    RelayState = auto()
    WaterTempCTimes1000 = auto()
    WaterTempFTimes1000 = auto()
    WaterFlowGpmTimes100 = auto()
    CurrentRmsMicroAmps = auto()
    GallonsPerMinuteTimes10 = auto()
    CurrentRmsMilliAmps = auto()
    VoltageRmsMilliVolts = auto()

    @classmethod
    def default(cls) -> "TelemetryName":
        """
        Returns default value Unknown
        """
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]



""" GwSchema Enums used in gwatn """
from gwatn.enums.actor_class import ActorClass
from gwatn.enums.algo_cert_type import AlgoCertType
from gwatn.enums.atn_spaceheat_strategy_name import AtnSpaceheatStrategyName
from gwatn.enums.core_g_node_role import CoreGNodeRole
from gwatn.enums.distribution_tariff import DistributionTariff
from gwatn.enums.emitter_pump_feedback_model import EmitterPumpFeedbackModel
from gwatn.enums.energy_supply_type import EnergySupplyType
from gwatn.enums.g_node_role import GNodeRole
from gwatn.enums.g_node_status import GNodeStatus
from gwatn.enums.gni_status import GniStatus
from gwatn.enums.local_comm_interface import LocalCommInterface
from gwatn.enums.make_model import MakeModel
from gwatn.enums.market_price_unit import MarketPriceUnit
from gwatn.enums.market_quantity_unit import MarketQuantityUnit
from gwatn.enums.market_type_name import MarketTypeName
from gwatn.enums.message_category import MessageCategory
from gwatn.enums.message_category_symbol import MessageCategorySymbol
from gwatn.enums.mixing_valve_feedback_model import MixingValveFeedbackModel
from gwatn.enums.recognized_currency_unit import RecognizedCurrencyUnit
from gwatn.enums.recognized_temperature_unit import RecognizedTemperatureUnit
from gwatn.enums.role import Role
from gwatn.enums.strategy_name import StrategyName
from gwatn.enums.supervisor_container_status import SupervisorContainerStatus
from gwatn.enums.telemetry_name import TelemetryName
from gwatn.enums.unit import Unit
from gwatn.enums.universe_type import UniverseType


__all__ = [
    "AlgoCertType",
    "AtnSpaceheatStrategyName",
    "CoreGNodeRole",
    "DistributionTariff",
    "EmitterPumpFeedbackModel",
    "EnergySupplyType",
    "GNodeRole",
    "GNodeStatus",
    "GniStatus",
    "LocalCommInterface",
    "MarketPriceUnit",
    "MarketQuantityUnit",
    "MarketTypeName",
    "MessageCategory",
    "MessageCategorySymbol",
    "MixingValveFeedbackModel",
    "RecognizedCurrencyUnit",
    "RecognizedTemperatureUnit",
    "ActorClass",
    "MakeModel",
    "Role",
    "TelemetryName",
    "Unit",
    "StrategyName",
    "SupervisorContainerStatus",
    "UniverseType",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gwatn',
 'gwatn.csv_makers',
 'gwatn.data_classes',
 'gwatn.dev_utils',
 'gwatn.enums',
 'gwatn.types']

package_data = \
{'': ['*'], 'gwatn': ['dispatch_contract_artifacts/*']}

install_requires = \
['boto3>=1.26.3,<2.0.0',
 'gridworks-protocol>=0.1.5,<0.2.0',
 'gridworks>=0.0.15,<0.0.16',
 'numpy>=1.23.4,<2.0.0',
 'paho-mqtt>=1.6.1,<2.0.0',
 'types-requests>=2.28.11.2,<3.0.0.0']

entry_points = \
{'console_scripts': ['gridworks-atn = gwatn.__main__:main']}

setup_kwargs = {
    'name': 'gridworks-atn',
    'version': '0.2.2',
    'description': 'Gridworks Atn Spaceheat',
    'long_description': "# Gridworks Atn\n\n[![PyPI](https://img.shields.io/pypi/v/gridworks-atn.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/gridworks-atn.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/gridworks-atn)][python version]\n[![License](https://img.shields.io/pypi/l/gridworks-atn)][license]\n\n[![Read the documentation at https://gridworks-atn.readthedocs.io/](https://img.shields.io/readthedocs/gridworks/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/thegridelectric/gridworks-atn/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/thegridelectric/gridworks-atn/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/gridworks-atn/\n[status]: https://pypi.org/project/gridworks-atn/\n[python version]: https://pypi.org/project/gridworks-atn\n[read the docs]: https://gridworks-atn.readthedocs.io/\n[tests]: https://github.com/thegridelectric/gridworks-atn/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/thegridelectric/gridworks-atn\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\nThis is the [GridWorks](https://gridworks.readthedocs.io/) Python SDK for building Atomic Transactive Nodes, or\n[AtomicTNodes](https://gridworks.readthedocs.io/en/latest/atomic-t-node.html>).\n\nAtomicTNodes are the most fun and interesting GridWorks actors to design and build. They are what make electrical devices\n_transactive_. More specifically, each AtomicTNode is dedicated to the job of operating its very own\n[Transactive Device](https://gridworks.readthedocs.io/en/latest/transactive-device.html), and simultaneously bidding on its behalf into electricity markets.\n\nTo learn about using this SDK, visit the [Gridworks Atn docs](https://gridworks-atn.readthedocs.io/en/latest/). To explore the rest of GridWorks, visit the [GridWorks docs](https://gridworks.readthedocs.io/en/latest/).\n\n## License\n\nDistributed under the terms of the [MIT license][license],\n**Gridworks Atn** is free and open source software.\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was generated from [@cjolowicz]'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/thegridelectric/gridworks-atn/issues\n[pip]: https://pip.pypa.io/\n\n<!-- github-only -->\n\n[license]: https://github.com/thegridelectric/gridworks-atn/blob/main/LICENSE\n[contributor guide]: https://github.com/thegridelectric/gridworks-atn/blob/main/CONTRIBUTING.md\n[command-line reference]: https://gridworks-atn.readthedocs.io/en/latest/usage.html\n",
    'author': 'Jessica Millar',
    'author_email': 'jmillar@gridworks-consulting.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/thegridelectric/gridworks-atn',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from fastapi import FastAPI
from fastapi import HTTPException
from gridworks.utils import RestfulResponse

from gwatn.dev_utils.dev_validator import DevValidator
from gwatn.types import TerminalassetCertifyHack


app = FastAPI()

validator = DevValidator()


@app.post("/terminalasset-certification/")
async def terminalasset_certify_hack_received(payload: TerminalassetCertifyHack):
    r = validator.terminalasset_certify_hack_received(payload)
    return r



"""Type scada.cert.transfer, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class ScadaCertTransfer(BaseModel):
    """Scada Certificate Transfer.

    This is a payload designed to be sent from a SCADA device to the GNodeFactory after the SCADA has opted into its certificate.
    """

    TaAlias: str = Field(
        title="TerminalAsset Alias",
        description="GNodeAlias of the TerminalAsset for which the SCADA certificate is issued. The ScadaCert can be found from this.",
    )
    SignedProof: str = Field(
        title="Signed Proof from the SCADA Actor",
        description="The Scada GNode has a ScadaAlgoAddr in the GNodeFactory database, and the identity of the SCADA actor can be verified by this.",
    )
    TypeName: Literal["scada.cert.transfer"] = "scada.cert.transfer"
    Version: str = "000"

    @validator("TaAlias")
    def _check_ta_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"TaAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("SignedProof")
    def _check_signed_proof(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedProof failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Scada is SignedProof signer.
        Axiom 1: Scada is SignedProof signer.
        There is a ScadaCert created by the Gnf with this ta_alias, and  the txn is the OptIn.
        """
        # TODO: Implement check for axiom 1"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class ScadaCertTransfer_Maker:
    type_name = "scada.cert.transfer"
    version = "000"

    def __init__(self, ta_alias: str, signed_proof: str):
        self.tuple = ScadaCertTransfer(
            TaAlias=ta_alias,
            SignedProof=signed_proof,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: ScadaCertTransfer) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> ScadaCertTransfer:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> ScadaCertTransfer:
        d2 = dict(d)
        if "TaAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaAlias")
        if "SignedProof" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedProof")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return ScadaCertTransfer(
            TaAlias=d2["TaAlias"],
            SignedProof=d2["SignedProof"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type gw.cert.id, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.enums import AlgoCertType
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


class AlgoCertType000SchemaEnum:
    enum_name: str = "algo.cert.type.000"
    symbols: List[str] = [
        "00000000",
        "086b5165",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class AlgoCertType000(StrEnum):
    ASA = auto()
    SmartSig = auto()

    @classmethod
    def default(cls) -> "AlgoCertType000":
        return cls.ASA

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class AlgoCertTypeMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> AlgoCertType:
        if not AlgoCertType000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to AlgoCertType000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, AlgoCertType, AlgoCertType.default())

    @classmethod
    def local_to_type(cls, algo_cert_type: AlgoCertType) -> str:
        if not isinstance(algo_cert_type, AlgoCertType):
            raise SchemaError(f"{algo_cert_type} must be of type {AlgoCertType}")
        versioned_enum = as_enum(
            algo_cert_type, AlgoCertType000, AlgoCertType000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, AlgoCertType000] = {
        "00000000": AlgoCertType000.ASA,
        "086b5165": AlgoCertType000.SmartSig,
    }

    versioned_enum_to_type_dict: Dict[AlgoCertType000, str] = {
        AlgoCertType000.ASA: "00000000",
        AlgoCertType000.SmartSig: "086b5165",
    }


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


class GwCertId(BaseModel):
    """Clarifies whether cert id is an Algorand Standard Asset or SmartSig"""

    Type: AlgoCertType = Field(
        title="Type",
    )
    Idx: Optional[int] = Field(
        title="ASA Index",
        default=None,
    )
    Addr: Optional[str] = Field(
        title="Algorand Smart Signature Address",
        default=None,
    )
    TypeName: Literal["gw.cert.id"] = "gw.cert.id"
    Version: str = "000"

    @validator("Type")
    def _check_type(cls, v: AlgoCertType) -> AlgoCertType:
        return as_enum(v, AlgoCertType, AlgoCertType.ASA)

    @validator("Addr")
    def _check_addr(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"Addr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Cert type consistency.
        If Type is ASA, then Id exists and Addr does not. Otherwise, Addr exists
        and Id does not.
        """
        Type: AlgoCertType = v.get("Type", None)
        Idx: int = v.get("Idx", None)
        Addr: str = v.get("Addr", None)

        if Type == AlgoCertType.ASA:
            if Idx is None:
                raise ValueError("If Type is ASA then Idx must exist")
            if Addr is not None:
                raise ValueError("If Type is ASA then Addr must be None")
        if Type == AlgoCertType.SmartSig:
            if Idx is not None:
                raise ValueError("If Type is SmartSig then Idx must be None")
            if Addr is None:
                raise ValueError("If Type is SmartSig then Addr must exist")
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Type"]
        Type = as_enum(self.Type, AlgoCertType, AlgoCertType.default())
        d["TypeGtEnumSymbol"] = AlgoCertTypeMap.local_to_type(Type)
        if d["Idx"] is None:
            del d["Idx"]
        if d["Addr"] is None:
            del d["Addr"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class GwCertId_Maker:
    type_name = "gw.cert.id"
    version = "000"

    def __init__(self, type: AlgoCertType, idx: Optional[int], addr: Optional[str]):
        self.tuple = GwCertId(
            Type=type,
            Idx=idx,
            Addr=addr,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: GwCertId) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> GwCertId:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GwCertId:
        d2 = dict(d)
        if "TypeGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeGtEnumSymbol")
        if d2["TypeGtEnumSymbol"] in AlgoCertType000SchemaEnum.symbols:
            d2["Type"] = AlgoCertTypeMap.type_to_local(d2["TypeGtEnumSymbol"])
        else:
            d2["Type"] = AlgoCertType.default()
        if "Idx" not in d2.keys():
            d2["Idx"] = None
        if "Addr" not in d2.keys():
            d2["Addr"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return GwCertId(
            Type=d2["Type"],
            Idx=d2["Idx"],
            Addr=d2["Addr"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type super.starter, version 000"""
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.types.g_node_instance_gt import GNodeInstanceGt
from gwatn.types.g_node_instance_gt import GNodeInstanceGt_Maker
from gwatn.types.supervisor_container_gt import SupervisorContainerGt
from gwatn.types.supervisor_container_gt import SupervisorContainerGt_Maker


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class SuperStarter(BaseModel):
    """Used by world to seed a docker container with data needed to spawn and superviser GNodeInstances"""

    SupervisorContainer: SupervisorContainerGt = Field(
        title="Key data about the docker container",
    )
    GniList: List[GNodeInstanceGt] = Field(
        title="List of GNodeInstances (Gnis) run in the container",
    )
    AliasWithKeyList: List[str] = Field(
        title="Aliases of Gnis that own Algorand secret keys",
    )
    KeyList: List[str] = Field(
        title="Algorand secret keys owned by Gnis",
    )
    TypeName: Literal["super.starter"] = "super.starter"
    Version: str = "000"

    @validator("GniList")
    def _check_gni_list(cls, v: List) -> List:
        for elt in v:
            if not isinstance(elt, GNodeInstanceGt):
                raise ValueError(
                    f"elt {elt} of GniList must have type GNodeInstanceGt."
                )
        return v

    @validator("AliasWithKeyList")
    def _check_alias_with_key_list(cls, v: List) -> List:
        for elt in v:
            try:
                check_is_left_right_dot(elt)
            except ValueError as e:
                raise ValueError(
                    f"AliasWithKeyList element {elt} failed LeftRightDot format validation: {e}"
                )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["SupervisorContainer"] = self.SupervisorContainer.as_dict()

        # Recursively call as_dict() for the SubTypes
        gni_list = []
        for elt in self.GniList:
            gni_list.append(elt.as_dict())
        d["GniList"] = gni_list
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class SuperStarter_Maker:
    type_name = "super.starter"
    version = "000"

    def __init__(
        self,
        supervisor_container: SupervisorContainerGt,
        gni_list: List[GNodeInstanceGt],
        alias_with_key_list: List[str],
        key_list: List[str],
    ):
        self.tuple = SuperStarter(
            SupervisorContainer=supervisor_container,
            GniList=gni_list,
            AliasWithKeyList=alias_with_key_list,
            KeyList=key_list,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: SuperStarter) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> SuperStarter:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SuperStarter:
        d2 = dict(d)
        if "SupervisorContainer" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainer")
        if not isinstance(d2["SupervisorContainer"], dict):
            raise SchemaError(
                f"d['SupervisorContainer'] {d2['SupervisorContainer']} must be a SupervisorContainerGt!"
            )
        supervisor_container = SupervisorContainerGt_Maker.dict_to_tuple(
            d2["SupervisorContainer"]
        )
        d2["SupervisorContainer"] = supervisor_container
        if "GniList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GniList")
        gni_list = []
        if not isinstance(d2["GniList"], List):
            raise SchemaError("GniList must be a List!")
        for elt in d2["GniList"]:
            if not isinstance(elt, dict):
                raise SchemaError(
                    f"elt {elt} of GniList must be "
                    "GNodeInstanceGt but not even a dict!"
                )
            gni_list.append(GNodeInstanceGt_Maker.dict_to_tuple(elt))
        d2["GniList"] = gni_list
        if "AliasWithKeyList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AliasWithKeyList")
        if "KeyList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing KeyList")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return SuperStarter(
            SupervisorContainer=d2["SupervisorContainer"],
            GniList=d2["GniList"],
            AliasWithKeyList=d2["AliasWithKeyList"],
            KeyList=d2["KeyList"],
            TypeName=d2["TypeName"],
            Version="000",
        )
